@@ -29,7 +29,7 @@ assert('include_graphics() includes custom images correctly', {
 hook_src = knit_hooks$get("source")
 options_ = list(engine = "r", prompt = FALSE, highlight = TRUE)
 
-assert('Attributes for souce can be specified class.source and attr.source', {
+assert('Attributes for source can be specified class.source and attr.source', {
   (hook_src("1", c(options_, class.source = "a b")) %==% "\n\n```{.r .a .b}\n1\n```\n\n")
   (hook_src("1", c(options_, attr.source = ".a .b")) %==% "\n\n```{.r .a .b}\n1\n```\n\n")
   (hook_src("1", c(options_, class.source = "a", attr.source = "b='1'")) %==%
@@ -38,6 +38,15 @@ assert('Attributes for souce can be specified class.source and attr.source', {
     "\n\n```{.r .a b='1'}\n1\n```\n\n")
 })
 
+assert('class.source and attr.source works also with collapse = TRUE', {
+  hook_chunk = hooks_markdown()$chunk
+  (hook_chunk("```{.r .a b=1}\n1\n```", c(options_, collapse = TRUE)) %==%
+      "```{.r .a b=1}\n1\n```")
+  (hook_chunk("```{.r .a b=1}\n1\n```\n```{.r .a b=1}\n1\n```",
+              c(options_, collapse = TRUE)) %==%
+      "```{.r .a b=1}\n1\n1\n```")
+})
+
 hook_out = knit_hooks$get("output")
 
 assert('Attributes for souce can be specified class.source and attr.source', {
